#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Fee views
"""

import datetime

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.security import checkPermission
from zope.security.checker import canAccess
from zope.traversing.browser.absoluteurl import absoluteURL

import zc.resourcelibrary
from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form import widget
from zc.table.column import GetterColumn

from schooltool import table
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.common import getResourceURL
from schooltool.common.inlinept import InheritTemplate
from schooltool.course.interfaces import ISectionContainer
from schooltool.group.interfaces import IGroupContainer
from schooltool.person.interfaces import IPerson
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.report.report import ReportLinkViewlet
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin import flourish
from schooltool.term.interfaces import IDateManager

from schooltool.fee.fee import Credit
from schooltool.fee.fee import Debit
from schooltool.fee.interfaces import IAmount
from schooltool.fee.interfaces import ICredit
from schooltool.fee.interfaces import IDebit
from schooltool.fee.interfaces import IFeeContainer
# XXX: remove this once the ARK changes get into core
from schooltool.fee.browser.section import is_student

from schooltool.fee import FeeMessage as _


def format_decimal(amount):
    return '%.2f' % amount


class BalanceRow(object):

    amount = 0
    title = _('Balance')
    description = ''

    def __init__(self):
        self.date = datetime.date.max

    @property
    def balance_amount(self):
        return self.amount


class FeePersonBase(object):

    @Lazy
    def person(self):
        return IPerson(self.context)


class PersonFeesViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/personfees.pt')

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    @Lazy
    def fees(self):
        return IFeeContainer(self.context)

    def render(self, *args, **kw):
        if is_student(self.context):
            zc.resourcelibrary.need('schooltool.fee.flourish')
            return self.template(*args, **kw)
        return ''


class AmountAddView(flourish.form.AddForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    @property
    def title(self):
        return self.person.title

    def updateActions(self):
        super(AmountAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(AmountAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'))
    def handleCancel(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return '%s?active_accordion=5' % absoluteURL(self.person, self.request)

    def create(self, data):
        obj = self.factory()
        form.applyChanges(self, obj, data)
        return obj

    def add(self, obj):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', obj)
        self.context[name] = obj


AddAmountDefaultDate = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.request.util.today,
    view=AmountAddView,
    field=IAmount['date']
    )


AddAmountDefaultAmount = widget.StaticWidgetAttribute(
    0,
    view=AmountAddView,
    field=IAmount['amount']
    )


field_order = ['date', 'title', 'description', 'amount']


class DebitAddView(AmountAddView):

    legend = _('Debit Information')
    fields = field.Fields(IDebit).select(*field_order)
    factory = Debit


class CreditAddView(AmountAddView):

    legend = _('Credit Information')
    fields = field.Fields(ICredit).select(*field_order)
    factory = Credit


class AmountEditView(flourish.form.Form, form.EditForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/amount_form.pt')

    @property
    def title(self):
        return self.person.title

    def update(self):
        form.EditForm.update(self)

    def updateActions(self):
        super(AmountEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        try:
            return self.request['camefrom']
        except (KeyError,):
            return '%s?active_accordion=5' % absoluteURL(self.person,
                                                         self.request)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        super(AmountEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class DebitEditView(AmountEditView):

    legend = _('Debit Information')
    fields = field.Fields(IDebit).select(*field_order)


class CreditEditView(AmountEditView):

    legend = _('Credit Information')
    fields = field.Fields(ICredit).select(*field_order)


class AmountActionsLinks(flourish.page.RefineLinksViewlet):

    pass


class AmountDeleteLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        return '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))


class FeeContainerDeleteView(flourish.containers.ContainerDeleteView,
                             FeePersonBase):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            return '%s?active_accordion=5' % absoluteURL(self.person,
                                                         self.request)
        return flourish.containers.ContainerDeleteView.nextURL(self)


def title_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return value
    return table.table.url_cell_formatter(value, item, formatter)


def amount_cell_formatter(value, item, formatter):
    return format_decimal(value)


def date_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return ''
    return table.table.url_cell_formatter(value, item, formatter)


def edit_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return ''
    template = '''
      <a class="modify" href="%(edit_url)s" title="%(link_title)s">
        <img src="%(edit_icon_url)s" alt="%(edit_title)s" />
      </a>'''
    request = formatter.request
    library = 'schooltool.skin.flourish'
    item_url = absoluteURL(item, request)
    person_url = absoluteURL(IPerson(item), request)
    return template % {
        'edit_url': '%s/edit.html?camefrom=%s' % (item_url, person_url),
        'link_title': translate(_('Edit this amount'), context=request),
        'edit_icon_url': getResourceURL(library, 'edit-icon.png', request),
        'edit_title': translate(STMessage('Edit'), context=request),
    }


class FeeDateColumn(table.column.DateColumn):

    pass


class FeeAmountColumn(GetterColumn):

    pass


class FeeContainerTable(table.ajax.Table):

    minimum_rows = 5

    @property
    def visible_column_names(self):
        result = ['action', 'date', 'title', 'amount']
        if not self.view.canModify:
            result = result[1:]
        return result

    def sortOn(self):
        return (('date', False), ('fee_type', False))

    def columns(self):
        actions = self.getActionColumns()
        date = FeeDateColumn(
            name='date',
            title=STMessage('Date'),
            getter=lambda i, f: i.date,
            cell_formatter=date_cell_formatter,
            subsort=True)
        fee_type = GetterColumn(
            name='fee_type',
            title=u'Fee type',
            getter=lambda i, f: ICredit.providedBy(i),
            subsort=True)
        title = GetterColumn(
            name='title',
            title=STMessage('Title'),
            getter=lambda i, f: i.title or '',
            cell_formatter=title_cell_formatter)
        amount = FeeAmountColumn(
            name='amount',
            title=_('Amount'),
            getter=lambda i, f: i.balance_amount,
            cell_formatter=amount_cell_formatter)
        return actions + [date, fee_type, title, amount]

    def getActionColumns(self):
        if self.view.canModify:
            return [
                GetterColumn(
                    name='action',
                    title=STMessage('Edit'),
                    getter=lambda i, f: i,
                    cell_formatter=edit_cell_formatter)
            ]
        return []

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data fees'})


class FeeContainerTableFilter(table.ajax.TableFilter):

    def filter(self, items):
        if items:
            items = list(items)
            balance = BalanceRow()
            balance.amount = sum([item.balance_amount for item in items])
            items.append(balance)
        return items

    def render(self, *args, **kw):
        return ''


class AmountFeesTableFilter(FeeContainerTableFilter):

    def filter(self, items):
        if items:
            items = list(items)
            balance = BalanceRow()
            balance.title = STMessage('Total')
            balance.amount = sum([item.amount for item in items])
            items.append(balance)
        return items


class AmountView(flourish.form.DisplayForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/amount.pt')
    fields = field.Fields(IAmount).select(*field_order).omit('title')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def title(self):
        return self.person.title

    def nextURL(self):
        return '%s?active_accordion=5' % absoluteURL(self.person, self.request)

    def edit_url(self):
        item_url = absoluteURL(self.context, self.request)
        return '%(url)s/edit.html?camefrom=%(url)s' % {'url': item_url}


class AmountDoneLink(flourish.viewlet.Viewlet):

    template = flourish.templates.Inline('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/nextURL"
         i18n:translate="">Done</a>
    </h3>
    ''')


class RequestAmountView(RequestRemoteReportDialog):

    report_builder = 'amount.pdf'


class RequestPersonFeeSummaryView(RequestRemoteReportDialog):

    report_builder = 'person_fee_summary.pdf'


class PersonFeeSummaryPDFView(flourish.report.PlainPDFPage, FeePersonBase):

    name = _('Fee Summary')
    message_title = _('fee summary report')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        dtm = getUtility(IDateManager)
        today = dtm.today
        return self.formatDate(today)

    @property
    def title(self):
        return self.person.title

    @property
    def base_filename(self):
        return '%s_fee_summary' % (self.context.username)


class ReceiptPDFView(PersonFeeSummaryPDFView):

    @property
    def name(self):
        if IPerson.providedBy(self.person):
            return _('Receipt')
        return _('Activity Credit')

    @property
    def message_title(self):
        if IPerson.providedBy(self.person):
            return _('receipt report')
        return _('activity credit report')

    @property
    def base_filename(self):
        return '%s_%s_receipt' % (self.person.__name__,
                                  self.context.date.strftime('%Y-%m-%d'))


class InvoicePDFView(PersonFeeSummaryPDFView):

    @property
    def name(self):
        if IPerson.providedBy(self.person):
            return _('Invoice')
        return _('Activity Debit')

    @property
    def message_title(self):
        if IPerson.providedBy(self.person):
            return _('invoice report')
        return _('activity debit report')

    @property
    def base_filename(self):
        return '%s_%s_invoice' % (self.person.__name__,
                                  self.context.date.strftime('%Y-%m-%d'))


class FeeSummaryPart(table.pdf.RMLTablePart):

    template = flourish.templates.XMLFile('rml/fee_summary_part.pt')
    table_name = "fees_table"

    def getColumnWidths(self, rml_columns):
        return '20%,65%,15%'

    def render(self):
        columns = self.getColumns()
        rml_columns = self.getRMLColumns(columns)

        if not rml_columns:
            return ''

        items = self.getItems()

        widths_string = self.getColumnWidths(rml_columns)
        headers = self.getHeaders(rml_columns)
        tables = []
        rows = []
        current_group = None
        for item in items:
            group = table.pdf.minimal_escape_rml(self.formatter.getSubGroup(item))
            if group != current_group:
                if rows:
                    tables.append({
                        'headers': current_group and [[current_group]] or [],
                        'rows': rows,
                        })
                current_group = group
                rows = []
            rows.append([column.renderCell(item, self.formatter)
                         for column in rml_columns])

        if rows:
            tables.append({
                    'headers': current_group and [[current_group]] or [],
                    'rows': rows,
                    })
        if tables and tables[-1].get('rows'):
            item_rows = len(tables[-1]['rows']) - 1 # balance row is included
            empty_rows_needed = self.table.minimum_rows - item_rows
            if empty_rows_needed > 0:
                for i in range(empty_rows_needed):
                    tables[-1]['rows'].insert(-1, ['']*len(rml_columns))
        rml = self.template(
            headers=headers, content=tables,
            col_widths=widths_string)
        return rml


class FeesTable(FeeContainerTable):

    visible_column_names = ['date', 'title', 'amount']

    @Lazy
    def source(self):
        return IFeeContainer(self.context)

    def getActionColumns(self):
        return []


class AmountFeesTable(FeesTable):

    @Lazy
    def source(self):
        return {self.context.__name__: self.context}

    def columns(self):
        date, fee_type, title, amount = super(AmountFeesTable, self).columns()
        amount.getter = lambda i, f: i.amount
        return [date, fee_type, title, amount]


class FeeRMLDateColumn(table.pdf.RMLDateColumn):

    def renderCell(self, item, formatter):
        if isinstance(item, BalanceRow):
            return ''
        return super(FeeRMLDateColumn, self).renderCell(item, formatter)


class FeeRMLAmountColumn(table.pdf.RMLGetterColumn):

    def renderCell(self, item, formatter):
        value = self.column.getter(item, formatter)
        return self.escape(format_decimal(value))


class FeeSummaryReportLinkViewlet(ReportLinkViewlet):

    def render(self, *args, **kw):
        if is_student(self.context):
            return self.template(*args, **kw)
        return ''


class InvoiceReportLink(ReportLinkViewlet):

    @property
    def title(self):
        if IPerson.providedBy(self.view.person):
            return _('Invoice')
        return _('Activity Debit')


class ReceiptReportLink(ReportLinkViewlet):

    @property
    def title(self):
        if IPerson.providedBy(self.view.person):
            return _('Receipt')
        return _('Activity Credit')


class RequestSchoolFeeSummaryView(RequestRemoteReportDialog):

    report_builder = 'fee_summary.pdf'


class SchoolFeeSummaryPDFView(flourish.report.PlainPDFPage):

    name = _('Fee Summary')
    message_title = _('school fee summary report')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        dtm = getUtility(IDateManager)
        today = dtm.today
        return self.formatDate(today)

    @property
    def title(self):
        return self.name

    @property
    def base_filename(self):
        return 'school_fee_summary'

    @Lazy
    def schoolyear_sections(self):
        result = []
        schoolyears = ISchoolYearContainer(self.context)
        active = schoolyears.getActiveSchoolYear()
        if active is not None:
            for term in active.values():
                result.extend(ISectionContainer(term).values())
            result.append(IGroupContainer(active)['students'])
        return result

    def is_student(self, sections, person):
        for section in sections:
            if person in section.members:
                return True
        return False

    def balance(self, person, date):
        fees = [fee for fee in IFeeContainer(person).values()
                if fee.date <= date]
        return sum([item.balance_amount for item in fees])

    def update(self):
        self.students = []
        self.paid_students = []
        self.not_paid_students = []
        dtm = getUtility(IDateManager)
        today = dtm.today
        for person in self.context['persons'].values():
            if self.is_student(self.schoolyear_sections, person):
                self.students.append(person)
                balance = self.balance(person, today)
                info = {'title': person.title, 'balance': balance}
                if balance >= 0:
                    self.paid_students.append(info)
                else:
                    self.not_paid_students.append(info)
        super(SchoolFeeSummaryPDFView, self).update()


class SchoolFeeSummaryHeaderPart(flourish.report.PDFPart):

    template = flourish.templates.XMLFile(
        'rml/school_fee_summary_header.pt')

    def paid_students(self):
        return '%d/%d' % (len(self.view.paid_students),
                          len(self.view.students))


class SchoolFeeSummaryTablePart(flourish.report.PDFPart):

    template = flourish.templates.XMLFile(
        'rml/school_fee_summary_table.pt')

    col_widths = "80%,20%"

    def update(self):
        collator = ICollator(self.request.locale)
        self.not_paid_students = sorted(self.view.not_paid_students,
                                        key=lambda i:collator.key(i['title']))
        super(SchoolFeeSummaryTablePart, self).update()


class SchoolFeeSummaryReportLink(ReportLinkViewlet,
                                 ActiveSchoolYearContentMixin):

    def render(self, *args, **kw):
        if self.schoolyear is None:
            return ''
        return super(SchoolFeeSummaryReportLink, self).render(*args, **kw)
